//#include "stdafx.h"
//#include "RVLCore2.h"

#include "PCLPointCloud.h"
#include <pcl/io/pcd_io.h>
#include <pcl/io/ply_io.h>
#include <pcl/surface/vtk_smoothing/vtk_utils.h>
#include <pcl/surface/organized_fast_mesh.h>
#include <pcl/surface/gp3.h>
#include <pcl/surface/poisson.h>
#include <pcl/filters/fast_bilateral.h>
#include <pcl/features/normal_3d_omp.h>
#include "RVLCore2.h"
#include "Util.h"
#include "RVLVTK.h"
#include "Graph.h"
#include "Mesh.h"
#include "RGBDCamera.h"
#include "PCLTools.h"
#include "PCLMeshBuilder.h"


using namespace RVL;

PCLMeshBuilder::PCLMeshBuilder()
{
	sigmaS = 5.0f;
	sigmaR = 0.05f;	
	normalEstR = 0.010f;
	organizedFastMeshMaxEdgeLen = -1.0f;
	voxelSize = 0.01f;
	nSDFFilter = 5;
	SDFIsoValue = 0.0f;
	flags = 0x00000000;

	vpBilateralFilter = new pcl::FastBilateralFilter<pcl::PointXYZRGBA>;
	vpNormalEstimator = new pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal>;
	vpOFM = new pcl::OrganizedFastMesh<pcl::PointXYZRGBA>;
}


PCLMeshBuilder::~PCLMeshBuilder()
{
	delete ((pcl::FastBilateralFilter<pcl::PointXYZRGBA> *)vpBilateralFilter);
	delete ((pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal> *)vpNormalEstimator);
	delete ((pcl::OrganizedFastMesh<pcl::PointXYZRGBA> *)vpOFM);
}

void PCLMeshBuilder::CreateMesh(
	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr PC,
	pcl::PolygonMesh &mesh)
{
	if (flags & RVLPCLMESHBUILDER_FLAG_ORGANIZED_PC)
	{
	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr PC_;

	int i;

	// Bilateral filtering

	if (flags & RVLPCLMESHBUILDER_FLAG_BILATERAL_FILTER)
	{
		pcl::FastBilateralFilter<pcl::PointXYZRGBA> *pBilateralFilter = (pcl::FastBilateralFilter<pcl::PointXYZRGBA> *)vpBilateralFilter;

		pBilateralFilter->setSigmaS((float)sigmaS);
		pBilateralFilter->setSigmaR((float)sigmaR);

		pBilateralFilter->setInputCloud(PC);

		pBilateralFilter->applyFilter(FPC);

		PC_ = { boost::make_shared<pcl::PointCloud<pcl::PointXYZRGBA>>(FPC) };
	}
	else
		PC_ = PC;

	// Compute normals

	pcl::search::OrganizedNeighbor<pcl::PointXYZRGBA>::Ptr tree(new pcl::search::OrganizedNeighbor<pcl::PointXYZRGBA>());

	pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal> *pNormalEstimator = (pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal> *)vpNormalEstimator;

	pNormalEstimator->setRadiusSearch((float)normalEstR);
	pNormalEstimator->setInputCloud(PC_);
	pNormalEstimator->setSearchMethod(tree);
	pNormalEstimator->compute(N);

	for (i = 0; i < N.points.size(); i++)
		if (!isfinite(N.points[i].normal_x))
		{
		N.points[i].normal_x = 0.0f;
		N.points[i].normal_y = 0.0f;
		N.points[i].normal_z = 0.0f;
		N.points[i].curvature = 0.0f;
		}


	// Create OrganizedFastMesh

	pcl::OrganizedFastMesh<pcl::PointXYZRGBA> *pOFM = (pcl::OrganizedFastMesh<pcl::PointXYZRGBA> *)vpOFM;

	pOFM->setTriangulationType(pcl::OrganizedFastMesh<pcl::PointXYZRGBA>::TRIANGLE_ADAPTIVE_CUT);

	pOFM->setInputCloud(PC_);

	if (organizedFastMeshMaxEdgeLen > 0.0f)
		pOFM->setMaxEdgeLength(organizedFastMeshMaxEdgeLen);
	
	pOFM->reconstruct(mesh);	// Conditions for adding a triangle are defined in function isShadowed in organized_fast_mesh.h.
	// This function is applied to endpoints of every triangle edge.
	// The meaning of most of the parameters of OrganizedFastMesh method can be understood from the code of this function.

	// Add normals to mesh

	//pcl::toPCLPointCloud2(*N, N2);
	pcl::toPCLPointCloud2(N, N2);	
		pcl::concatenateFields(N2, mesh.cloud, aux);
		mesh.cloud = aux;
	}
	else
	{
		pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal> *pNormalEstimator = (pcl::NormalEstimationOMP<pcl::PointXYZRGBA, pcl::Normal> *)vpNormalEstimator;

		pNormalEstimator->setRadiusSearch((float)normalEstR);
		pNormalEstimator->setInputCloud(PC);

		pcl::search::KdTree<pcl::PointXYZRGBA>::Ptr tree(new pcl::search::KdTree<pcl::PointXYZRGBA>());

		pNormalEstimator->setSearchMethod(tree);
		pNormalEstimator->compute(N);

		pcl::PCLPointCloud2 PCN2;
		pcl::toPCLPointCloud2(N, N2);
		pcl::PCLPointCloud2 PC2;
		pcl::toPCLPointCloud2(*PC, PC2);
		pcl::concatenateFields(PC2, N2, PCN2);

		pcl::PointCloud<pcl::PointNormal>::Ptr PCN(new pcl::PointCloud<pcl::PointNormal>());

		pcl::fromPCLPointCloud2(PCN2, *PCN);

		//print_info("Using parameters: depth %d, solverDivide %d, isoDivide %d\n", depth, solver_divide, iso_divide);

		pcl::Poisson<pcl::PointNormal> poisson;
		poisson.setDepth(8);
		poisson.setSolverDivide(8);
		poisson.setIsoDivide(8);
		poisson.setPointWeight(4.0f);
		poisson.setInputCloud(PCN);

		poisson.reconstruct(mesh);

		pcl::PointCloud<pcl::PointXYZRGBA>::Ptr meshPC(new pcl::PointCloud<pcl::PointXYZRGBA>());

		pcl::fromPCLPointCloud2(mesh.cloud, *meshPC);

		pNormalEstimator->setInputCloud(meshPC);
		pNormalEstimator->compute(N);
		pcl::toPCLPointCloud2(N, N2);

		pcl::concatenateFields(N2, mesh.cloud, aux);
		mesh.cloud = aux;
	}

	///
}

bool PCLMeshBuilder::CreateMesh(
	vtkSmartPointer<vtkPolyData> pPolygonData,
	pcl::PolygonMesh &mesh)
{
	vtkSmartPointer<vtkUnsignedCharArray> rgbPointData = rgbPointData->SafeDownCast(pPolygonData->GetPointData()->GetArray("Colors"));
	if (rgbPointData == NULL)
	{
		rgbPointData = rgbPointData->SafeDownCast(pPolygonData->GetPointData()->GetArray("RGB"));

		if (rgbPointData)
			rgbPointData->SetName("Colors");
		else
			return false;
	}

	pcl::VTKUtils::vtk2mesh(pPolygonData, mesh);

	vtkSmartPointer<vtkFloatArray> normalPointData = normalPointData->SafeDownCast(pPolygonData->GetPointData()->GetArray("Normals"));
	if (normalPointData == NULL)
	{
		normalPointData = normalPointData->SafeDownCast(pPolygonData->GetPointData()->GetNormals());

		if (normalPointData == NULL)
			return false;
	}

	int nPts = pPolygonData->GetNumberOfPoints();

	nPts = normalPointData->GetNumberOfTuples();

	N.resize(nPts);

	float N_[3];
	
	for (int i = 0; i < nPts; i++)
	{
		normalPointData->GetTupleValue(i, N_);

		N.points[i].normal_x = N_[0];
		N.points[i].normal_y = N_[1];
		N.points[i].normal_z = N_[2];
	}

	pcl::toPCLPointCloud2(N, N2);
	pcl::concatenateFields(N2, mesh.cloud, aux);
	mesh.cloud = aux;

	return true;
}

void PCLMeshBuilder::CreateParamList(CRVLMem *pMem)
{
	ParamList.m_pMem = pMem;

	RVLPARAM_DATA *pParamData;

	ParamList.Init();

	pParamData = ParamList.AddParam("MeshBuilder.sigmaS", RVLPARAM_TYPE_DOUBLE, &sigmaS);
	pParamData = ParamList.AddParam("MeshBuilder.sigmaR", RVLPARAM_TYPE_DOUBLE, &sigmaR);
	pParamData = ParamList.AddParam("MeshBuilder.normalEstR", RVLPARAM_TYPE_DOUBLE, &normalEstR);
	pParamData = ParamList.AddParam("MeshBuilder.organizedFastMesh.maxEdgeLen", RVLPARAM_TYPE_FLOAT, &organizedFastMeshMaxEdgeLen);
	pParamData = ParamList.AddParam("MeshBuilder.bilateralFilter", RVLPARAM_TYPE_FLAG, &flags);
	ParamList.AddID(pParamData, "yes", RVLPCLMESHBUILDER_FLAG_BILATERAL_FILTER);
	pParamData = ParamList.AddParam("MeshBuilder.organizedPC", RVLPARAM_TYPE_FLAG, &flags);
	ParamList.AddID(pParamData, "yes", RVLPCLMESHBUILDER_FLAG_ORGANIZED_PC);
	pParamData = ParamList.AddParam("MeshBuilder.width", RVLPARAM_TYPE_INT, &width);
	pParamData = ParamList.AddParam("MeshBuilder.height", RVLPARAM_TYPE_INT, &height);
	pParamData = ParamList.AddParam("MeshBuilder.voxelSize", RVLPARAM_TYPE_FLOAT, &voxelSize);
	pParamData = ParamList.AddParam("MeshBuilder.nSDFFilter", RVLPARAM_TYPE_INT, &nSDFFilter);
	pParamData = ParamList.AddParam("MeshBuilder.SDFIsoValue", RVLPARAM_TYPE_FLOAT, &SDFIsoValue);
}

bool PCLMeshBuilder::Load(
	char *FileName,
	Mesh *pMesh,
	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr PC_,
	pcl::PolygonMesh &PCLMesh_,
	bool bSavePLY)
{
	char *fileExtension = RVLGETFILEEXTENSION(FileName);

	if (strcmp(fileExtension, "ply") == 0)
	{
		if (flags & RVLPCLMESHBUILDER_FLAG_VISIBLE_SURFACE)
		{
			vtkSmartPointer<vtkPLYReader> reader = vtkSmartPointer<vtkPLYReader>::New();
			reader->SetFileName(FileName);
			reader->Update();
			vtkSmartPointer<vtkPolyData> pPolygonData = reader->GetOutput();

			printf("Creating visible surface mesh...");

			pMesh->pPolygonData = MESH::CreateVisibleSurfaceMesh(pPolygonData, voxelSize, nSDFFilter, SDFIsoValue);

			printf("completed.\n");

			pMesh->bOrganizedPC = false;

			char *PLYFileName = RVLCreateFileName(FileName, ".ply", -1, "_vs.ply");

			printf("Saving mesh to %s...", PLYFileName);

			pMesh->SavePolyDataToPLY(PLYFileName);

			printf("completed.\n");

			delete[] PLYFileName;

			return false;
		}
		else
			pMesh->LoadPolyDataFromPLY(FileName);
	}
	else
	{
		if (strcmp(fileExtension, "pcd") == 0)
			PCLLoadPCD(FileName, PC_);
		else if (strcmp(fileExtension, "bmp") == 0)
		{
			IplImage *RGBImage = cvLoadImage(FileName);

			char *depthFileName = RVLCreateString(FileName);

			Array2D<short int> depthImage;

			depthImage.Element = NULL;
			depthImage.w = depthImage.h = 0;

			sprintf(RVLGETFILEEXTENSION(depthFileName), "txt");

			unsigned int format;

			if (ImportDisparityImage(depthFileName, depthImage, format))
			{				
				printf("Creating point cloud from RGB-D image.\n");

				camera.GetPointCloud(&depthImage, RGBImage, PC_);

				RVL_DELETE_ARRAY(depthImage.Element);
			}
			else
			{
				sprintf(RVLGETFILEEXTENSION(depthFileName), "png");

				cv::Mat depthImagePNG = cv::imread(depthFileName, CV_LOAD_IMAGE_ANYDEPTH);

				depthImage.w = depthImagePNG.cols;
				depthImage.h = depthImagePNG.rows;

				depthImage.Element = (short int *)(depthImagePNG.data);

				camera.GetPointCloud(&depthImage, RGBImage, PC_);
			}

			delete[] depthFileName;
			cvReleaseImage(&RGBImage);
		}
		else
		{
			printf("ERROR: Unknown file format!\n");

			return false;
		}

		printf("Creating organized PCL mesh from point cloud...");

		CreateMesh(PC_, PCLMesh_);

		printf("completed.\n");

		if (bSavePLY)
		{
			char *PLYFileName = RVLCreateString(FileName);

			sprintf(RVLGETFILEEXTENSION(PLYFileName), "ply");

			printf("Saving mesh to %s...", PLYFileName);

			PCLSavePLY(PLYFileName, PCLMesh_);

			printf("completed.\n");

			delete[] PLYFileName;
		}

		vtkSmartPointer<vtkPolyData> pd = vtkSmartPointer<vtkPolyData>::New();
		PCLMeshToPolygonData(PCLMesh_, pd);
		pMesh->pPolygonData = vtkSmartPointer<vtkPolyData>::New();
		pMesh->pPolygonData->DeepCopy(pd);
	}

	if (flags & RVLPCLMESHBUILDER_FLAG_ORGANIZED_PC)
	{
		pMesh->bOrganizedPC = true;
		pMesh->width = width;
		pMesh->height = height;
	}
	else
		pMesh->bOrganizedPC = false;

	//commented 28.12.2017 - Vidovic (time measurement TPAMI18)
	//printf("Creating ordered mesh from PCL mesh...");

	pMesh->CreateOrderedMeshFromPolyData();

	//commented 28.12.2017 - Vidovic (time measurement TPAMI18)
	//printf("completed.\n");

	return true;
}

void RVL::CreateMesh(
	void *vpMeshBuilder,
	Array2D<short int> *pDepthImage,
	IplImage *pRGBImage,
	Mesh *pMesh,
	bool bSavePLY,
	char *FileName)
{
	PCLMeshBuilder *pMeshBuilder = (PCLMeshBuilder *)vpMeshBuilder;

	pMeshBuilder->camera.GetPointCloud(pDepthImage, pRGBImage, pMeshBuilder->PC);

	pMeshBuilder->CreateMesh(pMeshBuilder->PC, pMeshBuilder->PCLMesh);

	if (bSavePLY)
	{
		char *PLYFileName = RVLCreateString(FileName);

		sprintf(RVLGETFILEEXTENSION(PLYFileName), "ply");

		printf("Saving mesh to %s...", PLYFileName);

		PCLSavePLY(PLYFileName, pMeshBuilder->PCLMesh);

		printf("completed.\n");

		delete[] PLYFileName;
	}

	vtkSmartPointer<vtkPolyData> pd = vtkSmartPointer<vtkPolyData>::New();
	PCLMeshToPolygonData(pMeshBuilder->PCLMesh, pd);
	pMesh->pPolygonData = vtkSmartPointer<vtkPolyData>::New();
	pMesh->pPolygonData->DeepCopy(pd);

	pMesh->bOrganizedPC = true;
	pMesh->width = pDepthImage->w;
	pMesh->height = pDepthImage->h;

	printf("Creating ordered mesh from PCL mesh...");

	pMesh->CreateOrderedMeshFromPolyData();

	printf("completed.\n");
}

bool RVL::LoadMesh(
	void *vpMeshBuilder,
	char *FileName,
	Mesh *pMesh,
	bool bSavePLY)
{
	PCLMeshBuilder *pMeshBuilder = (PCLMeshBuilder *)vpMeshBuilder;

	return pMeshBuilder->Load(FileName, pMesh, pMeshBuilder->PC, pMeshBuilder->PCLMesh, bSavePLY);
}

